#include "ninox.h"

// Command parser:
// "var" or "var=value"
int
SetVariable(char *str)
	{
	char cmdbuffer[1024],fname[1024];
	int i,j,bval;
	char *arg = cmdbuffer;
	char *val = cmdbuffer;

	/* Find the value */
	strcpy(cmdbuffer,str);
	while(*val && *val != '=') ++val;
	if (*val == '=') { 
	   *(val++) = 0;
	   while (*val==' ') ++val; 
	   }

	bval=1;
	if (!strcasecmp(val,"on") || 	!strcasecmp(val,"yes") || !strcasecmp(val,"true") ||
					!strcasecmp(val,"1"))
		bval=1;
	else if (!strcasecmp(val,"off") || 	!strcasecmp(val,"no") || !strcasecmp(val,"false") ||
						!strcasecmp(val,"0"))
		bval=0;

	if (!strcasecmp(arg,"inputfilter")) {
		if (bval==0) {InputFilter=0; return 1;}

		if (InputFilter==0 && atof(val) > 0)
		   InputFilter_ThreshHold = atof(val);
		++InputFilter;
		return 1;
		}

	if (!strcasecmp(arg,"i")) {
		InteractiveMode=bval;
		return 1;
		}

	if (!strcasecmp(arg,"enablecutout") || !strcasecmp(arg,"cutout")) {
		DoCutout=bval;
		return 1;
		}
	if (!strcasecmp(arg,"detectbrokenframes") || !strcasecmp(arg,"dbf")) {
		if (!strcasecmp(val,"planet")) {
		   DBF_TYPE = DBF_PLANET;
		   bval=1;
		   }
		if (!strcasecmp(val,"lunar")) {
		   DBF_TYPE = DBF_LUNAR;
		   bval=1;
		   }

		DetectBrokenFrames = bval;
		return 1;
	   	}

	if (!strcasecmp(arg,"dbfedgecount")) {
		DBF_EDGE_COUNT = atoi(val);
		return 1;
		}

	if (!strcasecmp(arg,"detecthotpixels") || !strcasecmp(arg,"dhp")) {
		DetectHotPixels = bval;
		return 1;
	   	}

	// give a non-central place to start scanning for FindCentre()
	if (!strcasecmp(arg,"scan")) {
	   int x1,y1;
	   char fname[512];

	   if (sscanf(val,"%d,%d",&x1,&y1) == 2) {
		Sub_x = x1;
		Sub_y = y1;
	 	return 1;
		}
	   printf("-scan: invalid format, should be -scan=x,y\n");
	   exit(1);
	   }

	// subregion specifies a region to be used after the image has been centered
	// and cutout. This subregion replaces the centered image
	if (!strcasecmp(arg,"subregion")) {
	   int x1,y1,x2,y2;
	   char fname[512];

	   if (sscanf(val,"%d,%d,%d,%d",&x1,&y1,&x2,&y2) == 4) {
		EnableSubRegion = 1;
		SR_X1 = x1;
		SR_X2 = x2;
		SR_Y1 = y1;
		SR_Y2 = y2;

		DoCutout=0;
		return 1;
	 	}

	   printf("-subregion: Invalid format, should be -subregion=x1,y1,x2,y2\n");
	   exit(1);
	   }

	if (!strcasecmp(arg,"darkframe")) {
		double scale;
		char fname[512];

		if (sscanf(val,"%lf,%s",&scale,fname) == 2) {
		   DarkFrame = strdup(fname);
		   DarkFrame_Scale = scale;
		   }
		else {
		   DarkFrame = strdup(val);
		   DarkFrame_Scale = 1.0;
		   }

		return 1;
	   	}

	if (!strcasecmp(arg,"levels")) {
		int min,max;
		if (sscanf(val,"%d,%d",&min,&max) == 2) {
		   LevelsMin = min;
		   LevelsMax = max;
		   doLevelsAdjust = 1;
		   return 1;
		   }
		printf("syntax error on 'levels'\n");
		return 0;
	   	}

	if (!strcasecmp(arg,"dbf-diff")) {
		DBF_DIFF = atoi(val);
		return 1;
	   	}

	if (!strcasecmp(arg,"nocutout")) {
		DoCutout=0;
		return 1;
		}

	if (!strcasecmp(arg,"invert")) {
		doInvertImage = bval;
		return 1;
		}

	if (!strcasecmp(arg,"forceprocess")) {
		ForceProcess = bval;
		return 1;
		}

	if (!strcasecmp(arg,"protect")) {
		HistoProtect = bval;  // Don't prevent histo-stretching
		return 1;
		}

	if (!strcasecmp(arg,"stretch")) {
		HistoStretch = atoi(val);
		if (HistoStretch==0) HistoStretch = 200;
		return 1;
		}

	if (!strcasecmp(arg,"istretch")) {
		InputHistoStretch = 1;
		return 1;
		}

	if (!strcasecmp(arg,"noistretch")) {
		InputHistoStretch = 0;
		return 1;
		}

	if (!strcasecmp(arg,"noprotect")) {
		HistoProtect = 0;  // Don't protect
		return 1;
		}

	if (!strcasecmp(arg,"quiet")) {
		Quiet = bval;
		return 1;
		}

	if (!strcasecmp(arg,"detectbrokenframes")) {
		DetectBrokenFrames = 1;
		return 1;
		}

	if (!strcasecmp(arg,"debayer")) {
		WantDebayer = bval;
		return 1;
		}

	if (!strcasecmp(arg,"cutx")) {
		int i = atoi(val);
		if (i < 100) {
		   Print("-cutx: minimum supported size 100 pixels\n");
		   }
		else CutX = i;
		return 1;
		}

	if (!strcasecmp(arg,"cuty")) {
		int i = atoi(val);
		if (i < 100) {
		   Print("-cutx: minimum supported size 100 pixels\n");
		   }
		else CutY = i;
		return 1;
		}

	if (!strcasecmp(arg,"cut")) {
		int x,y;

		if (sscanf(val,"%d,%d",&x,&y)==2) {
		   CutX = x;
		   CutY = y;
		   }
		else {
		   int i = atoi(val);
		   if (i < 100 || i>1000) {
		      	Print("%d: Outside valid bounds 100..1000\n",i);
		      	return 0;
			}
		   CutX = i; CutY = i;
		   }

		newWidth = CutX;
		newHeight = CutY;
		return 1;
		}

	if (!strcasecmp(arg,"white")) {
		White = atoi(val);
		return 1;
		}

	if (!strcasecmp(arg,"smoothing")) {
		if (!strcasecmp(val,"before")) {
		   UpScale_Smoothing_When = SMOOTHING_BEFORE;
		   UpScale_Smoothing = 1;
		   }
		else if (!strcasecmp(val,"after")) {
		   UpScale_Smoothing_When = SMOOTHING_AFTER;
		   UpScale_Smoothing = 1;
		   }
		else UpScale_Smoothing = bval;
		return 1;
		}

	if (!strcasecmp(arg,"nosmoothing")) {
		UpScale_Smoothing = 0;
		return 1;
		}

	if (!strcasecmp(arg,"upscale") || !strcasecmp(arg,"upsample") ||
		!strcasecmp(arg,"resample") || !strcasecmp(arg,"rescale")) {
		char *ptr = strstr(val,"/");
		int up=1,down=1;

		if (ptr) {
		   *ptr = 0;
		   down = atoi(ptr+1);
		   }

		up = atoi(val);
		if (up < 1 || up > 6 || down<1 || down>6) {
		   Print("UpScale: %d/%d out of accepted range (1..6)/(1..6) \n",up,down);
		   return 1;
		   }
		UpScale = up;
		DownScale = down;

	 	if (UpScale_Smoothing < 0) {
                   double scalef = (double)UpScale / (double) DownScale;
                   if (scalef < 2) UpScale_Smoothing = 0;
		   else UpScale_Smoothing = 1;
		   }

		return 1;
		}

	if (!strcasecmp(arg,"popfilter")) {
	 	char when[32];
		int sz = 5;

		if (sscanf(val,"%d,%s",&sz,when) == 2) {
		   if (!strcasecmp(when,"before")) PopFilter_When = POPFILTER_BEFORE;
		   else if (!strcasecmp(when,"after")) PopFilter_When = POPFILTER_AFTER;
		   else {
			Print("Unknown option '%s' to popfilter\n",when);
			return(0);
			}
		   if (sz<=0) {
			Print("invalid size arg '%d' to popfilter\n",sz);
			return(0);
			}
		   PopFilter = sz;
		   }
		else if (! strcasecmp(val,"before")) {
			PopFilter_When = POPFILTER_BEFORE;
			PopFilter = 5;
			}
		else if (! strcasecmp(val,"after")) {
			PopFilter_When = POPFILTER_AFTER;
			PopFilter = 5;
			}
		else if (atoi(val)>0) {
			PopFilter = atoi(val);
			PopFilter_When = POPFILTER_BEFORE;
			}
		else {
			PopFilter = 5;
			PopFilter_When = POPFILTER_BEFORE;
			}

		return 1;
		}

	if (!strcasecmp(arg,"gamma")) {
		double g = atof(val);
		if (g < 0.3 || g > 1.5) {
		   Print("Gamma value %f is outside accepted range 0.3 <= gamma <= 1.5\n",g);
		   return 1;
		   }
		Gamma = g;
		if (Gamma != 1.0) ChangeGamma = 1;
		return 1;
		}

	if (!strcasecmp(arg,"streamfilter")) {
		StreamFilter = 1;
		if (atoi(val)) {
		   StreamFilter_History = atoi(val);
		   }
		return 1;
		}

	if (!strcasecmp(arg,"bayerboost")) {
		BayerBoost = bval;
		return 1;
		}

	if (!strcasecmp(arg,"overwrite")) {
		AllowOverwrite = bval;
		return 1;
		}

	if (!strcasecmp(arg,"outdir")) {
		OutDir = strdup(val);
		for(j=0; OutDir[j]; ++j) if (OutDir[j]=='\\') OutDir[j]='/';
		return 1;
		}

	if (!strcasecmp(arg,"nosave")) {
		NoSave = bval;
		return 1;
		}

	if (!strcasecmp(arg,"threshhold")) {
		int t = atoi(val);
		if (t<5 || t > 250) {
		   Print("Threshhold %d out of allowed range 5..250\n",t);
		   return 1;
		   }
		ThreshHold = t;
		return 1;
		}

	if (!strcasecmp(arg,"minpixels")) {
		int m = atoi(val);
		if (m<10 || m > 1000) {
		   Print("Minpixels: %d out of alloed range 10..1000\n",m);
		   }
		else MinPixels = m;
		return 1;
		}

	if (!strcasecmp(arg,"width")) {
		int w = (int)atoi(val);
		if (w < 100 || w>1000)
		   Print("Width must be 100..1000\n");
		else newWidth = w;
		return 1;
		}

	if (!strcasecmp(arg,"height")) {
		int h = (int)atoi(val);
		if (h < 100 || h > 1000)
		   Print("Height must be 100..1000\n");
		else newHeight = h;
		return 1;
		}

	if (!strcasecmp(arg,"stack") || !strcasecmp(arg,"stackfile")) {
		StackFile = strdup(val);
		return 1;
		}

	if (!strcasecmp(arg,"stackcount")) {
		StackMax = atoi(val);
		return 1;
		}

	if (!strcasecmp(arg,"mergefile")) {
		MergeFile = strdup(val);
		return 1;
		}

	if (!strcasecmp(arg,"mergethreshhold")) {
		MergeThreshHold = atof(val);
		return 1;
		}

	if (!strcasecmp(arg,"writeemptyfiles")) {
		forceWriteEmptyFiles = bval;
		WriteEmptyFiles = bval;
		return 1;
		}

	if (!strcasecmp(arg,"qestimator")) {
		int min,max;
		QEstimator = bval;
		// Possible value "n,m" representing min and max subsampling
		if (sscanf(val,"%d,%d",&min,&max)==2) {
		   QSUBSAMPLE_MIN = min;
		   QSUBSAMPLE_MAX = max;
		   QEstimator = 1;
		   }
		return 1;
		}

	if (!strncasecmp(arg,"qfunc",4)) {
	   if (! strncasecmp(val,"hist",4)) QualityFunction = HISTO;
	   else if (! strncasecmp(val,"grad",4)) QualityFunction = GRADIENT; // default
	   else {
		printf("Unknown quality function '%s'\n",val);
		return 0;
		}
	   return 1;
	   }

	if (!strcasecmp(arg,"qhistfile")) {
	   QHISTO_FILE = strdup(val);
	   return 1;
	   }

	if (!strncasecmp(arg,"qhisto",6)) {
	   int x1,y1,x2,y2,min,max;
	   if (sscanf(val,"%d,%d-%d,%d:%d-%d",&x1,&y1,&x2,&y2,&min,&max)==6)
		QAddRegion(x1,y1,x2,y2,min,max);
	   if (sscanf(val,"%d,%d",&min,&max)==2) {
	   	QHISTO_MIN = min;
		QHISTO_MAX = max;
		}
	   else {
		printf("Usage: -qhisto=min,max (8 bit values)\n");
		return 0;
		}
	   return 1;
	   }

	if (!strcasecmp(arg,"renumber") || !strcasecmp(arg,"qrenumber") || !strcasecmp(arg,"qsort")) {
	   	if (bval==0)
		   QRenumberFiles = 0;
		else if (!strcasecmp(val,"first")) QRenumberFiles = QRENUMBER_FIRST;
		else if (!strcasecmp(val,"last")) QRenumberFiles = QRENUMBER_LAST;
		else QRenumberFiles = QRENUMBER_FIRST; // default
		return 1;
		}

	if (!strcasecmp(arg,"qwrite")) {
		QWriteIntermediateFiles = bval;
		return 1;
		}

	if (!strcasecmp(arg,"hnoisefilter")) {
		HNoiseFilter = bval;
		return 1;
		}

	if (!strcasecmp(arg,"subdirs")) {
	   SubDirs = bval;
	   return 1;
	   }

	if (!strcasecmp(arg,"nosubdirs")) {
	   SubDirs = 0;
	   return 1;
	   }

	if (!strcasecmp(arg,"status")) {
		Status();
		exit(1);
		}

	if (!strcasecmp(arg,"debug")) {
		DEBUG = bval;
		return 1;
		}

	// Support for multipoint alignment and image morphing
	if (!strcasecmp(arg,"morphing")) {
		Morphing = atoi(val);
		return 1;
		}

	if (!strcasecmp(arg,"morph_ref") || !strcasecmp(arg,"morph-ref")) {
		Morph_Ref = strdup(val);
		if (access(Morph_Ref,0)) {
		   Print("morph_ref: Reference image '%s' does not exist\n",val);
		   exit(1);
		   }
		return 1;
	 	}
	if (!strcasecmp(arg,"morph_across") || !strcasecmp(arg,"morph-across")) {
		// Number of control points across the image
		Morph_Across = atoi(val);
		return 1;
	 	}

	if (!strncasecmp(arg,"morph_it",8) || !strncasecmp(arg,"morph-it",8)) {
		// Number of morph iterations to run
		Morph_Iterations = atoi(val);
		return 1;
	 	}

	if (!strncasecmp(arg,"morph_dr",8) || !strncasecmp(arg,"morph-dr",8)) {
		// 
		Morph_Drift = atoi(val);
		return 1;
	 	}

	if (!strncasecmp(arg,"morph_de",8) || !strncasecmp(arg,"morph-de",8)) {
		// Enable extra debugging
		Morph_Debug = atoi(val);
		return 1;
	 	}

	if (!strcasecmp(arg,"output-type")) {
		if (!strcasecmp(val,"bmp")) 		OutputFileType = IMG_BMP;
		else if (!strncasecmp(val,"fit",3)) 	OutputFileType = IMG_FIT;
		else if (!strcasecmp(val,"ppm")) 	OutputFileType = IMG_PPM;
		else {
		   Print("output file type '%s' not understood\n",val);
		   return 0;
		   }
	  	return 1;
		}

	if (!strcasecmp(arg,"output-depth")) {
	   OutputFileDepth = atoi(val);
	   return 1;
	   }

	if (!strcasecmp(arg,"rotate")) {
	   RotateAngle = atof(val);
	   if (RotateAngle != 0) do_Rotate=1;
	   return 1;
	   }

	if (!strcasecmp(arg,"rotatecutoff")) {
	   RotateCutoff = atoi(val);
	   Print("RotateCutoff = %d\n",RotateCutoff);
	   return 1;
	   }

       if (!strcasecmp(arg,"gaincomp")) {
               ApplyGainComp = 1;

               if (! LoadGainCompRef(val)) {
                  Print("Error loading gain compensation reference\n");
                  exit(1);
                  }

               return 1;
               }

       // Special test mode for simulated sampling
       if (!strcasecmp(arg,"subsamplemode")) {
               SubsampleMode = atoi(val);
               NoSave = 1;
               return 1;
               }

       if (!strcasecmp(arg,"Subsample_pixel_width"))  {Subsample_pixel_width = atoi(val); return 1;}
       if (!strcasecmp(arg,"Subsample_pixel_height")) {Subsample_pixel_height = atoi(val); return 1;}
       if (!strcasecmp(arg,"Subsample_pixel_size"))   {
               Subsample_pixel_height = Subsample_pixel_width = atoi(val); return 1;
               }
       if (!strcasecmp(arg,"Subsample_xpixels"))      {Subsample_xpixels= atoi(val); return 1;}
       if (!strcasecmp(arg,"Subsample_ypixels"))      {Subsample_ypixels= atoi(val); return 1;}
       if (!strcasecmp(arg,"Subsample_nframes"))      {Subsample_nframes = atoi(val); return 1;}
       if (!strcasecmp(arg,"Subsample_min_offset"))   {Subsample_min_offset = atof(val); return 1;}
       if (!strcasecmp(arg,"Subsample_max_offset"))   {Subsample_max_offset = atof(val); return 1;}
       if (!strcasecmp(arg,"Subsample_upscale"))      {Subsample_upscale = atoi(val); return 1;}
       if (!strcasecmp(arg,"Subsample_smooth"))       {Subsample_smooth = atoi(val); return 1;}
       if (!strcasecmp(arg,"Subsample_noise"))        {Subsample_noise = atoi(val); return 1;}
       if (!strcasecmp(arg,"Subsample_scan"))        {Subsample_scan = bval; return 1;}
       if (!strcasecmp(arg,"Subsample_prefix"))        {Subsample_prefix = strdup(val); return 1;}

	// Otherwise we don't recognise this variable
	return 0;
	}

int
Status()
	{
	int i;

	printf("Variables:\n");
	printf("Debug = %d\n",DEBUG);
	printf("CurrentFile = \"%s\"\n",CurrentFile?CurrentFile:"");

	printf("InteractiveMode = %d\n",InteractiveMode);
	printf("cutout = %d\n",DoCutout);
	printf("ForceProcess = %d\n",ForceProcess);
	printf("width = %d\n",newWidth);
	printf("height = %d\n",newHeight);
	printf("White = %d\n",White);
	printf("overwrite = %d\n",AllowOverwrite);
	printf("protect = %d\n",HistoProtect);
	printf("UpScale = %d\n",UpScale);
	printf("DownScale = %d\n",DownScale);
	printf("UpScale_Smoothing = %d\n",UpScale_Smoothing);
	printf("ThreshHold = %d\n",ThreshHold);
	printf("MinPixels = %d\n",MinPixels);
	printf("debayer = %d\n",WantDebayer);
	printf("InputFilter = %d\n",InputFilter);
	printf("InputFilter_ThreshHold = %lf\n",InputFilter_ThreshHold);
	printf("ChangeGamma = %d\n",ChangeGamma);
	printf("Gamma = %lf\n",Gamma);
	printf("CutX = %d\n",CutX);
	printf("CutY = %d\n",CutY);
	printf("BayerBoost = %lf\n",BayerBoost);
	printf("NoSave = %d\n",NoSave);
	printf("OutDir = \"%s\"\n",OutDir?OutDir:"");
	printf("WriteEmptyFiles = %d\n",WriteEmptyFiles);
	printf("Quiet = %d\n",Quiet);
	printf("StackCount = %d\n",StackCount);
	printf("StackMax = %d\n",StackMax);
	printf("StackFile = \"%s\"\n",StackFile?StackFile:"");
	printf("MergeFile = \"%s\"\n",MergeFile?MergeFile:"");
	printf("MergeThreshHold = %lf\n",MergeThreshHold);
	printf("forceWriteEmptyFiles = %d\n",forceWriteEmptyFiles);
	printf("HNoiseFilter = %d\n",HNoiseFilter);

	printf("qwrite = %d\n",QWriteIntermediateFiles);
	printf("renumber = %d\n",QRenumberFiles);
	printf("qestimator = %d\n",QEstimator);
	printf("QSUBSAMPLE_MIN = %d\n",QSUBSAMPLE_MIN);
	printf("QSUBSAMPLE_MAX = %d\n",QSUBSAMPLE_MAX);
	}

char *
ReadCommand(char *buffer, int size)
	{
	char *ptr;

	*buffer=0;
	fgets(buffer,size-1,stdin);

	// Strip off leading whitespace
	while(*buffer && *buffer <= 32) ++buffer;
	
	// Strip off trailing whitespace
	ptr = buffer + strlen(buffer);
	while(ptr != buffer && *ptr<=32) *(ptr--)=0;
	
	return buffer;
	}

int
Interact(void)
	{
	char *cmd,buffer[1024];
	Print("Interactive Mode\n");

	while(1) {
	   
	   cmd = ReadCommand(buffer,1024);
	   if (!strcasecmp(cmd,"quit") || !strcasecmp(cmd,"exit"))
		return(1);

	   if (!strncasecmp(cmd,"set ",4)) {
		SetVariable(cmd+4);
		continue;
		}

	   if (!strcasecmp(cmd,"status")) {
		Status();
		continue;
		}
	   }
	
	// notreached
	return 0;
	}
